/****************************************************************************
 *
 *    Copyright (c) 2020 Vivante Corporation
 *
 *    Permission is hereby granted, free of charge, to any person obtaining a
 *    copy of this software and associated documentation files (the "Software"),
 *    to deal in the Software without restriction, including without limitation
 *    the rights to use, copy, modify, merge, publish, distribute, sublicense,
 *    and/or sell copies of the Software, and to permit persons to whom the
 *    Software is furnished to do so, subject to the following conditions:
 *
 *    The above copyright notice and this permission notice shall be included in
 *    all copies or substantial portions of the Software.
 *
 *    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 *    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 *    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 *    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 *    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 *    FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 *    DEALINGS IN THE SOFTWARE.
 *
 *****************************************************************************/

#include "VsiPreparedModel.h"

#include <sys/mman.h>
#include <sys/system_properties.h>

#include "OperationsUtils.h"
#include "VsiDriver.h"
#if ANDROID_SDK_VERSION >= 28
#include "ValidateHal.h"
#endif
#include <nnrt/error.hpp>
#include <nnrt/event.hpp>

namespace android {
namespace nn {
namespace vsi_driver {

class VsiDriver;

static nnrt::OperationType op_code_mapping(
    HalPlatform::OperationType op) {  // Android O 8.1 API LEVEL 27
    switch (op) {
#define MAP_OP(op)                             \
    case HalPlatform::OperationType::op: {     \
        LOG(INFO) << "add operation: " << #op; \
        return nnrt::OperationType::op;        \
    }
        MAP_OP(ADD);
        MAP_OP(CONV_2D);
        MAP_OP(DEPTHWISE_CONV_2D);
        MAP_OP(RELU);
        MAP_OP(RESHAPE);
        MAP_OP(FULLY_CONNECTED);
        MAP_OP(SOFTMAX);
        MAP_OP(CONCATENATION);
        MAP_OP(AVERAGE_POOL_2D);
        MAP_OP(MAX_POOL_2D);
        MAP_OP(MUL);
        MAP_OP(RELU1);
        MAP_OP(RELU6);
        MAP_OP(TANH);
        MAP_OP(LOGISTIC);
        MAP_OP(FLOOR);
        MAP_OP(DEQUANTIZE);
        MAP_OP(SPACE_TO_DEPTH);
        MAP_OP(DEPTH_TO_SPACE);
        MAP_OP(L2_NORMALIZATION);
        MAP_OP(RESIZE_BILINEAR);
        MAP_OP(LOCAL_RESPONSE_NORMALIZATION);
        MAP_OP(EMBEDDING_LOOKUP);
        MAP_OP(RNN);
        MAP_OP(HASHTABLE_LOOKUP);
        MAP_OP(LSTM);
        MAP_OP(SVDF);
        MAP_OP(LSH_PROJECTION);
        MAP_OP(L2_POOL_2D);
#if ANDROID_SDK_VERSION >= 28
        MAP_OP(BATCH_TO_SPACE_ND);
        MAP_OP(DIV);
        MAP_OP(MEAN);
        MAP_OP(PAD);
        MAP_OP(SPACE_TO_BATCH_ND);
        MAP_OP(SQUEEZE);
        MAP_OP(STRIDED_SLICE);
        MAP_OP(SUB);
        MAP_OP(TRANSPOSE);
#endif
#if ANDROID_SDK_VERSION >= 29
        MAP_OP(ABS);
        MAP_OP(RESIZE_NEAREST_NEIGHBOR);
        MAP_OP(ARGMAX);
        MAP_OP(ARGMIN);
        MAP_OP(MAXIMUM);
        MAP_OP(MINIMUM);
        MAP_OP(SQRT);
        MAP_OP(RSQRT);
        MAP_OP(LOG);
        MAP_OP(EXP);
        MAP_OP(SIN);
        MAP_OP(REDUCE_MAX);
        MAP_OP(REDUCE_MIN);
        MAP_OP(REDUCE_PROD);
        MAP_OP(REDUCE_SUM);
        MAP_OP(REDUCE_ALL);
        MAP_OP(REDUCE_ANY);
        MAP_OP(NEG);
        MAP_OP(PRELU);
        MAP_OP(UNIDIRECTIONAL_SEQUENCE_RNN);
        MAP_OP(BIDIRECTIONAL_SEQUENCE_RNN);
        MAP_OP(UNIDIRECTIONAL_SEQUENCE_LSTM);
        MAP_OP(BIDIRECTIONAL_SEQUENCE_LSTM);
        MAP_OP(GENERATE_PROPOSALS);
        MAP_OP(AXIS_ALIGNED_BBOX_TRANSFORM);
        MAP_OP(DETECTION_POSTPROCESSING);
        MAP_OP(LESS);
        MAP_OP(LESS_EQUAL);
        MAP_OP(EQUAL);
        MAP_OP(GREATER);
        MAP_OP(GREATER_EQUAL);
        MAP_OP(NOT_EQUAL);
        MAP_OP(LOGICAL_AND);
        MAP_OP(LOGICAL_OR);
        MAP_OP(LOGICAL_NOT);
        MAP_OP(EXPAND_DIMS);
        MAP_OP(POW);
        MAP_OP(INSTANCE_NORMALIZATION);
        MAP_OP(SPLIT);
        MAP_OP(LOG_SOFTMAX);
        MAP_OP(GATHER);
        MAP_OP(CAST);
        MAP_OP(HEATMAP_MAX_KEYPOINT);
        MAP_OP(QUANTIZE);
        MAP_OP(RANDOM_MULTINOMIAL);
        MAP_OP(SELECT);
        MAP_OP(ROI_ALIGN);
        MAP_OP(GROUPED_CONV_2D);
        MAP_OP(CHANNEL_SHUFFLE);
        MAP_OP(TRANSPOSE_CONV_2D);
        MAP_OP(SLICE);
        case HalPlatform::OperationType::TOPK_V2: {
            LOG(INFO) << "add operation: TOPK_V2";
            return nnrt::OperationType::TOPK;
        };
#endif
#if ANDROID_SDK_VERSION >= 30
        MAP_OP(HARD_SWISH);
        MAP_OP(ELU);
#endif
#undef MAP_OP

        default:
            LOG(ERROR) << "Unknown operation code:" << static_cast<int32_t>(op);
            break;
    }

    return nnrt::OperationType::NONE;
};

static nnrt::OperandType operand_mapping(HalPlatform::OperandType code) {
#define MAP_OPERAND(code)                      \
    case HalPlatform::OperandType::code: {     \
        LOG(INFO) << "add operand: " << #code; \
        return nnrt::OperandType::code;        \
    }

    switch (code) {
#if ANDROID_SDK_VERSION >= 29
        MAP_OPERAND(BOOL);
        MAP_OPERAND(FLOAT16);
        MAP_OPERAND(TENSOR_BOOL8);
        MAP_OPERAND(TENSOR_FLOAT16);
        MAP_OPERAND(TENSOR_QUANT8_SYMM);
        MAP_OPERAND(TENSOR_QUANT16_ASYMM);
        MAP_OPERAND(TENSOR_QUANT16_SYMM);
        MAP_OPERAND(TENSOR_QUANT8_SYMM_PER_CHANNEL);
#endif
#if ANDROID_SDK_VERSION >= 30
    case HalPlatform::OperandType::TENSOR_QUANT8_ASYMM_SIGNED: {
        LOG(INFO) << "add operand: TENSOR_QUANT8_ASYMM_SIGNED";
        return nnrt::OperandType::TENSOR_QUANT8_ASYMM_SIGNED;
    }
#endif
        MAP_OPERAND(FLOAT32);
        MAP_OPERAND(INT32);
        MAP_OPERAND(UINT32);
        MAP_OPERAND(TENSOR_FLOAT32);
        MAP_OPERAND(TENSOR_INT32);
        MAP_OPERAND(TENSOR_QUANT8_ASYMM);
        default:
            break;
    }

#undef MAP_OPERAND

    return nnrt::OperandType::NONE;
}

SandBox VsiPreparedModel::sandbox_;

void VsiPreparedModel::release_rtinfo(std::vector<VsiRTInfo>& rtInfos) {
    while (!rtInfos.empty()) {
        rtInfos.pop_back();
    }
}

// #if ANDROID_SDK_VERSION < 30
Return<ErrorStatus> VsiPreparedModel::map_rtinfo_from_hidl_memory(
    const hidl_vec<hidl_memory>& pools, std::vector<VsiRTInfo>& rtInfos) {
    rtInfos.clear();
    rtInfos.resize(pools.size());

    for (size_t i = 0; i < pools.size(); i++) {
        if (!mapHidlMem(pools[i], rtInfos[i])) {
            return ErrorStatus::INVALID_ARGUMENT;
        }
    }
    return ErrorStatus::NONE;
}

#if ANDROID_SDK_VERSION < 30
void VsiPreparedModel::fill_operand_value(nnrt::op::OperandPtr ovx_operand,
                                          const HalPlatform::Operand& hal_operand) {
    switch (hal_operand.lifetime) {
        case OperandLifeTime::MODEL_INPUT:
        case OperandLifeTime::MODEL_OUTPUT:
        case OperandLifeTime::TEMPORARY_VARIABLE:
            // Skip lifetime is TEMPORARY_VARIABLE, MODEL_INPUT, MODEL_OUTPUT, or NO_VALUE
            break;
        case OperandLifeTime::NO_VALUE: {
            native_model_->setOperandValue(ovx_operand, nullptr, 0);
        } break;
        case OperandLifeTime::CONSTANT_COPY: {
            const auto& location = hal_operand.location;
            native_model_->setOperandValue(
                ovx_operand, model_.operandValues.data() + location.offset, location.length);
        } break;
        case OperandLifeTime::CONSTANT_REFERENCE: {
            const auto& location = hal_operand.location;
            auto& rt_info = const_buffer_[location.poolIndex];

            if ("ashmem" == rt_info.mem_type || "hardware_buffer_blob" == rt_info.mem_type) {
                const uint8_t* buffer = rt_info.ptr;
                native_model_->setOperandValue(
                    ovx_operand, buffer + location.offset, location.length);
            } else if ("mmap_fd" == rt_info.mem_type) {
                native_model_->setOperandValueFromMemory(
                    ovx_operand, rt_info.vsi_mem.get(), location.offset, location.length);
            }
        } break;
    }
}
#elif ANDROID_SDK_VERSION >= 30
void VsiPreparedModel::fill_operand_value(nnrt::op::OperandPtr ovx_operand,
                                          const HalPlatform::Operand& hal_operand) {
    switch (hal_operand.lifetime) {
        case OperandLifeTime::SUBGRAPH_INPUT:
        case OperandLifeTime::SUBGRAPH_OUTPUT:
        case OperandLifeTime::TEMPORARY_VARIABLE:
        case OperandLifeTime::SUBGRAPH:
            // Skip lifetime is TEMPORARY_VARIABLE, SUBGRAPH_INPUT, SUBGRAPH_OUTPUT, NO_VALUE
            // ,SUBGRAPH
            break;
        case OperandLifeTime::NO_VALUE: {
            native_model_->setOperandValue(ovx_operand, nullptr, 0);
        } break;
        case OperandLifeTime::CONSTANT_COPY: {
            const auto& location = hal_operand.location;
            native_model_->setOperandValue(
                ovx_operand, model_.operandValues.data() + location.offset, location.length);
        } break;
        case OperandLifeTime::CONSTANT_REFERENCE: {
            const auto& location = hal_operand.location;
            auto& rt_info = const_buffer_[location.poolIndex];

            if ("ashmem" == rt_info.mem_type || "hardware_buffer_blob" == rt_info.mem_type) {
                const uint8_t* buffer = rt_info.ptr;
                native_model_->setOperandValue(
                    ovx_operand, buffer + location.offset, location.length);
            } else if ("mmap_fd" == rt_info.mem_type) {
                native_model_->setOperandValueFromMemory(
                    ovx_operand, rt_info.vsi_mem.get(), location.offset, location.length);
            }
        } break;
    }
}
#endif

Return<ErrorStatus> VsiPreparedModel::construct_ovx_operand(
    nnrt::op::OperandPtr ovx_operand, const HalPlatform::Operand& hal_operand) {
    auto ovx_nnrt_type = operand_mapping(hal_operand.type);
    if (nnrt::OperandType::NONE == ovx_nnrt_type) {
        LOG(ERROR) << " do not support operand type: " << static_cast<int>(hal_operand.type);
        return ErrorStatus::INVALID_ARGUMENT;
    }
    ovx_operand->type = ovx_nnrt_type;
    ovx_operand->quant.scalar.scale = hal_operand.scale;
    ovx_operand->quant.scalar.zeroPoint = hal_operand.zeroPoint;
    ovx_operand->dimensions = hal_operand.dimensions;
#if ANDROID_SDK_VERSION >= 29
    if (ovx_nnrt_type == nnrt::OperandType::TENSOR_QUANT8_SYMM_PER_CHANNEL) {
        auto& channelQuant = hal_operand.extraParams.channelQuant();
        ovx_operand->quant.vec.channelDim = channelQuant.channelDim;
        ovx_operand->quant.vec.scale.resize(channelQuant.scales.size());
        ovx_operand->quant.vec.zeroPoint.resize(channelQuant.scales.size(), 0);
        for (size_t i = 0; i < channelQuant.scales.size(); i++) {
            ovx_operand->quant.vec.scale[i] = channelQuant.scales[i];
        }
    }
#endif
    // TODO: add check error
    switch (ovx_operand->type) {
        case nnrt::OperandType::FLOAT32:
        case nnrt::OperandType::INT32:
        case nnrt::OperandType::UINT32:
            break;
        case nnrt::OperandType::TENSOR_FLOAT32:
        case nnrt::OperandType::TENSOR_INT32:
        case nnrt::OperandType::TENSOR_QUANT8_ASYMM: {
            break;
        }
        default:
            break;
    }
    return ErrorStatus::NONE;
}

Return<ErrorStatus> VsiPreparedModel::initialize() {
    // [0] validate HAL::Model, return ErrorCode if validate failed
    // For scalar operand, dimension must be 0
    // [1] create async procedure to prepare model
    // [1.0] convert HAL model to nnrt::Model
    LOG(INFO) << "VsiPreparedModel::initialize()";

    auto status = map_rtinfo_from_hidl_memory(model_.pools, const_buffer_);
    if (ErrorStatus::NONE != status) return status;

    if(native_model_->get_cache_size() != 0 && native_model_->get_cache_handle() != -1){
        return initializeCacheInternel();
    }
    // add operand and set its value
#if ANDROID_SDK_VERSION < 30
    for (const auto& hal_operand : model_.operands) {
#elif ANDROID_SDK_VERSION >= 30
    for (const auto& hal_operand : model_.main.operands) {
#endif
        uint32_t registered_idx = 0;
        auto ovx_operand = native_model_->addOperand(nullptr, &registered_idx);
        auto status = construct_ovx_operand(ovx_operand, hal_operand);
        if (ErrorStatus::NONE != status) {
            LOG(INFO) << "Deivce do not support the operand type"
                      << static_cast<int32_t>(hal_operand.type);
            return status;
        }
        fill_operand_value(ovx_operand, hal_operand);
    }

    std::string not_supported_reason;
#if ANDROID_SDK_VERSION < 30
    for (const auto& hal_op : model_.operations) {
#elif ANDROID_SDK_VERSION >= 30
    for (const auto& hal_op : model_.main.operations) {
#endif
        if (IsOpBlocked(static_cast<int32_t>(hal_op.type)) ||
            !VsiDriver::isSupportedOperation(hal_op, model_, not_supported_reason)) {
            LOG(ERROR) << "Device do not support operation type: " << toString(hal_op.type);
            LOG(INFO) << not_supported_reason;
            return ErrorStatus::INVALID_ARGUMENT;
        }

        auto ovx_op_type = op_code_mapping(hal_op.type);
        if (nnrt::OperationType::NONE == ovx_op_type) {
            LOG(ERROR) << "Device do not support operation type: " << toString(hal_op.type);
            return ErrorStatus::INVALID_ARGUMENT;
        }

        auto ovx_op = native_model_->addOperation(ovx_op_type /* Operation Type*/,
                                                  &hal_op.inputs[0],    /*inputs */
                                                  hal_op.inputs.size(), /*num of inputs */
                                                  &hal_op.outputs[0],   /*outputs */
                                                  hal_op.outputs.size() /*num of outputs */
        );

        ovx_op->setDataLayout(nnrt::DataLayout::NHWC);
    }

    if (model_.relaxComputationFloat32toFloat16)
        native_model_->relax(true);  // convert fp32 data to fp16 in nnrt.

    native_model_->finish();
#if ANDROID_SDK_VERSION < 30
    std::vector<uint32_t> inputs = model_.inputIndexes;
    std::vector<uint32_t> outputs = model_.outputIndexes;
#elif ANDROID_SDK_VERSION >= 30
    std::vector<uint32_t> inputs = model_.main.inputIndexes;
    std::vector<uint32_t> outputs = model_.main.outputIndexes;
#endif
    native_model_->identifyInputsAndOutputs(
        inputs.data(), inputs.size(), outputs.data(), outputs.size());
    return ErrorStatus::NONE;
}

Return<ErrorStatus> VsiPreparedModel::initializeCacheInternel() {
    LOG(INFO) << "VsiPreparedModel::initializeCache()";

#if ANDROID_SDK_VERSION < 30
    std::vector<uint32_t> inputs = model_.inputIndexes;
    std::vector<uint32_t> outputs = model_.outputIndexes;
#elif ANDROID_SDK_VERSION >= 30
    std::vector<uint32_t> inputs = model_.main.inputIndexes;
    std::vector<uint32_t> outputs = model_.main.outputIndexes;
#endif

    auto op = std::make_shared<nnrt::op::Operation>(nnrt::OperationType::NBG);
    op->setInputs(inputs);
    op->setOutputs(outputs);
    op->setDataLayout(nnrt::DataLayout::NHWC);
    uint32_t out_index = 0;
    native_model_->addOperation(op, &out_index);

#if ANDROID_SDK_VERSION < 30
    for (const auto& hal_operand : model_.operands) {
#elif ANDROID_SDK_VERSION >= 30
    for (const auto& hal_operand : model_.main.operands) {
#endif
        uint32_t registered_idx = 0;
        auto ovx_operand = native_model_->addOperand(nullptr, &registered_idx);
        if (std::find(inputs.begin(), inputs.end(), registered_idx) != inputs.end() ||
            std::find(outputs.begin(), outputs.end(), registered_idx) != outputs.end() )
        {
            auto status = construct_ovx_operand(ovx_operand, hal_operand);
            if (ErrorStatus::NONE != status) {
                LOG(INFO) << "Deivce do not support the operand type"
                          << static_cast<int32_t>(hal_operand.type);
                return status;
            }
            fill_operand_value(ovx_operand, hal_operand);
        }
    }

    if (model_.relaxComputationFloat32toFloat16) native_model_->relax(true);

    native_model_->finish();
    native_model_->identifyInputsAndOutputs(
        inputs.data(), inputs.size(), outputs.data(), outputs.size());
    return native_model_->allocate_cache_memory(native_model_->get_cache_size())
               ? ErrorStatus::NONE
               : ErrorStatus::INVALID_ARGUMENT;
}

void VsiPreparedModel::update_operand_from_request(const std::vector<uint32_t>& indexes,
                                                   const hidl_vec<RequestArgument>& arguments) {
    nnAssert(indexes.size() == arguments.size());
    auto ovx_operands = native_model_->getOperands(indexes);
    for (size_t i = 0; i < indexes.size(); i++) {
        const RequestArgument& request_arg = arguments[i];
        auto& ovx_operand = ovx_operands[i];
        if (request_arg.dimensions.size() > 0) {
            // It's the responsibility of the caller to validate that
            // arg.dimensions only modifies the dimensions that were
            // unspecified in the model.  That's the case in SampleDriver.cpp
            // with the call to validateRequest().
            // TODO: after revaluing the dimension, model should be re-compiled
            ovx_operand->dimensions = request_arg.dimensions;
        }
    }
}

// #if ANDROID_SDK_VERSION < 30
Return<ErrorStatus> VsiPreparedModel::update_pool_info_from_request(
    const Request& request,
    const std::vector<uint32_t>& indexes,
    const hidl_vec<RequestArgument>& arguments,
    IO flag,
    std::vector<OutputShape>& outputShapes) {
    nnAssert(indexes.size() == arguments.size());
    auto ovx_operands = native_model_->getOperands(indexes);
    for (size_t i = 0; i < indexes.size(); i++) {
        const RequestArgument& request_arg = arguments[i];
        auto ovx_operand = ovx_operands[i];
        if (request_arg.hasNoValue) {
            if (flag == IO::INPUT)
                native_exec_->setInput(i, nullptr /*operand */, nullptr, 0);
            else
                native_exec_->setOutput(i, nullptr, nullptr, 0);
        } else {
            auto location = request_arg.location;
            auto poolIndex = location.poolIndex;
            nnAssert(poolIndex < request.pools.size());

            if (flag == IO::OUTPUT) {
                outputShapes[i].dimensions = ovx_operand->dimensions;
                outputShapes[i].isSufficient =
                    io_buffer_[poolIndex].buffer_size >= location.length + location.offset;
                if (io_buffer_[poolIndex].buffer_size < location.length + location.offset) {
                    LOG(ERROR) << "No Sufficient output, Execute failed";
                    return ErrorStatus::OUTPUT_INSUFFICIENT_SIZE;
                }
            }

            auto& rt_info = io_buffer_[poolIndex];
            int nn_return = NNA_NO_ERROR;
            if ("ashmem" == rt_info.mem_type || "hardware_buffer_blob" == rt_info.mem_type) {
                uint8_t* buffer = rt_info.ptr;
                if (flag == IO::INPUT)
                    nn_return = native_exec_->setInput(
                        i, ovx_operand, buffer + location.offset, location.length);
                else {
                    nn_return = native_exec_->setOutput(
                        i, ovx_operand, buffer + location.offset, location.length);
                }
            } else if ("mmap_fd" == rt_info.mem_type) {
                auto& vsi_mem = rt_info.vsi_mem;
                if (flag == IO::INPUT)
                    nn_return = native_exec_->setInputFromMemory(
                        i, ovx_operand, vsi_mem.get(), location.offset, location.length);
                else
                    nn_return = native_exec_->setOutputFromMemory(
                        i, ovx_operand, vsi_mem.get(), location.offset, location.length);
            } else if ("token" == rt_info.mem_type) {
                if (flag == IO::INPUT) {
                    nn_return = native_exec_->setInput(
                        i, ovx_operand, rt_info.ptr, rt_info.buffer_size);
                } else {
                    nn_return = native_exec_->setOutput(
                        i, ovx_operand, rt_info.ptr, rt_info.buffer_size);
                }
            } else if ("hardware_buffer" == rt_info.mem_type) {
                LOG(ERROR) << "Input & Ouput not set-1";
            } else {
                LOG(ERROR) << "Input & Ouput not set.";
            }

            auto status = convertResultCodeToErrorStatus(nn_return);
            if (status != ErrorStatus::NONE) {
                LOG(ERROR) << "fail to set IO, return error code :";
                return status;
            }
        }
    }

    return ErrorStatus::NONE;
}

Return<ErrorStatus> VsiPreparedModel::convertResultCodeToErrorStatus(int resultCode) {
    switch (resultCode) {
        case NNA_NO_ERROR:
            return ErrorStatus::NONE;

        case NNA_BAD_DATA:
        case NNA_UNEXPECTED_NULL:
            LOG(ERROR) << "INVALID_ARGUMENT";
            return ErrorStatus::INVALID_ARGUMENT;

        case NNA_OUTPUT_INSUFFICIENT_SIZE:
            LOG(ERROR) << "output insufficient size";
            return ErrorStatus::OUTPUT_INSUFFICIENT_SIZE;

        case NNA_BAD_STATE:
        case NNA_INCOMPLETE:
        case NNA_OP_FAILED:
        case NNA_OUT_OF_MEMORY:
        case NNA_UNMAPPABLE:
            LOG(ERROR) << "GENERAL_FAILURE";
            return ErrorStatus::GENERAL_FAILURE;

        default:
            LOG(ERROR) << "Unknown result code " << resultCode
                       << " mapped to ErrorStatus::GENERAL_FAILURE";
            return ErrorStatus::GENERAL_FAILURE;
    }
}

template <typename T_IExecutionCallback>
Return<ErrorStatus> VsiPreparedModel::executeBase(const Request& request,
                                                  const T_IExecutionCallback& callback) {
    LOG(INFO) << __FUNCTION__;

#if ANDROID_SDK_VERSION >= 30
    if (!validateRequest(request, HalPlatform::convertToV1_0(model_))) {
#else
    if (!validateRequest(request, model_)) {
#endif
        LOG(ERROR) << "invalid request";
        callback->notify(ErrorStatus::INVALID_ARGUMENT);
        return ErrorStatus::INVALID_ARGUMENT;
    }
    if (nullptr == native_compile_) {
        // native_compile_ = std::make_shared<nnrt::Compilation>(native_model_.get());
        native_compile_.reset(new nnrt::Compilation(native_model_.get()),
                              NativeDeleter<nnrt::Compilation>());
    }
    map_rtinfo_from_hidl_memory(request.pools, io_buffer_);
    if (!native_exec_) {
        // native_exec_ = std::make_shared<nnrt::Execution>(native_compile_.get());
        native_exec_.reset(new nnrt::Execution(native_compile_.get()),
                           NativeDeleter<nnrt::Execution>());
    }

    std::vector<hidl_memory> io_pools = request.pools;
    std::vector<RequestArgument> input_args = request.inputs;
    std::vector<RequestArgument> output_args = request.outputs;
    std::vector<OutputShape> outputShapes(request.outputs.size());

#if ANDROID_SDK_VERSION < 30
    update_operand_from_request(model_.inputIndexes, input_args);
    update_operand_from_request(model_.outputIndexes, output_args);
#elif ANDROID_SDK_VERSION >= 30
    update_operand_from_request(model_.main.inputIndexes, input_args);
    update_operand_from_request(model_.main.outputIndexes, output_args);
#endif
    if (!native_model_->isCompiled()) {
        native_compile_->run();
    }
#if ANDROID_SDK_VERSION < 30
    auto status = update_pool_info_from_request(
        request, model_.inputIndexes, input_args, IO::INPUT, outputShapes);
#elif ANDROID_SDK_VERSION >= 30
    auto status = update_pool_info_from_request(
        request, model_.main.inputIndexes, input_args, IO::INPUT, outputShapes);
#endif
    if (ErrorStatus::NONE != status) {
        callback->notify(status);
        return status;
    }
#if ANDROID_SDK_VERSION < 30
    status = update_pool_info_from_request(
        request, model_.outputIndexes, output_args, IO::OUTPUT, outputShapes);
#elif ANDROID_SDK_VERSION >= 30
    status = update_pool_info_from_request(
        request, model_.main.outputIndexes, output_args, IO::OUTPUT, outputShapes);
#endif
    if (ErrorStatus::NONE != status) {
        callback->notify(status);
        return status;
    }

    int error = native_exec_->compute();
    status = convertResultCodeToErrorStatus(error);
    release_rtinfo(io_buffer_);

    if (status != ErrorStatus::NONE) {
        callback->notify(ErrorStatus::INVALID_ARGUMENT);
        return status;
    }

    callback->notify(ErrorStatus::NONE);
    return ErrorStatus::NONE;
}

Return<ErrorStatus> VsiPreparedModel::execute(const Request& request,
                                              const sp<V1_0::IExecutionCallback>& callback) {
    ErrorStatus err;
    VsiPreparedModel::sandbox_.wait([this, &request, &callback, &err]() -> bool {
        err = this->executeBase(request, callback);
        return false;
    });

    native_exec_.reset();
    return err;
}

}  // namespace vsi_driver
}  // namespace nn
}  // namespace android
